/* maf_long_rows.c -- convert an ordered reference-sequence maf file to a few
*    long rows of nucleotides and dashes.  The default output format is that
*    used for text alignments by MultiPipMaker; appending the command-line
*    argument "fasta" requests FastA-format output.
*/

#include "util.h"
#include "maf.h"
#include "seq.h"

struct edge {
	struct name_rec *name;
	struct edge *next;
};

#define WHITE 0
#define GRAY 1
#define BLACK 2
struct name_rec {
	char *name;
	int color;
	struct edge *follows;
	struct name_rec *next;
};

char **row_name;
int next_row_name;

// depth-first search
static void dfs(struct name_rec *n) {
	struct edge *e;

	if (n->color != WHITE)
		fatal("please apply the maf_order program");
	n->color = GRAY;
	for (e = n->follows; e != NULL; e = e->next)
		if (e->name->color != BLACK)
			dfs(e->name);
	if (next_row_name <= 0)
		fatal("underflow in row names");
	row_name[--next_row_name] = copy_string(n->name);
	n->color = BLACK;
}

// compare reference row of block with the reference sequence
static void check_ref(struct mafComp *c, uchar *seq, int seq_len, int ncol) {
	char *s, x;
	int pos, col;

	pos = c->start;
	s = c->text;
	for(col = 0; col < ncol; ++col)
		if ((x = s[col]) != '-') {
			if (pos >= seq_len || toupper(x) != toupper(seq[pos]))
				fatalf("ref-seq mismatch at position %d", pos);
			++pos;
		}
}

int main(int argc, char **argv)
{
	struct mafFile *mf;
	struct mafAli *a;
	struct mafComp *c, *d;
	SEQ *sf;
	uchar *s;
	char **row;
	struct name_rec *names, *m, *n;
	struct edge *e;
	int i, j, seq_len, nrow, ncol, next_pos, col, fasta = 0;

	argv0 = "maf_long_rows";
	if (argc != 3 && argc != 4)
		fatal("args = reference-sequence-file maf-file [fasta]");
	
	if (argc == 4) {
		if (same_string(argv[3], "fasta"))
			fasta = 1;
		else
			fatal("third argument must be \"fasta\"");
	}
	sf = seq_get(argv[1]);
	s = SEQ_CHARS(sf);
	seq_len = SEQ_LEN(sf);
	mf = mafReadAll(argv[2], 0);
	next_pos = nrow = ncol = 0;
	names = NULL;
	for (a = mf->alignments; a != NULL; a = a->next) {
		c = a->components;
		if (c->start < next_pos)
			fatalf("alignments out of order at pos %d", c->start);
		check_ref(c, s, seq_len, a->textSize);
		ncol += (c->start - next_pos + a->textSize);
		next_pos = c->start + c->size;
		if (nrow == 0) {
			names = ckalloc(sizeof(struct name_rec));
			names->name = copy_string(c->src);
			names->follows = NULL;
			names->next = NULL;
			nrow = 1;
		} else if (!same_string(c->src,names->name))
			fatalf("conflicting ref-seq names: %s and %s",
			  names->name, c->src);
		for (m = names ; (d = c->next) != NULL; c = d, m = n) {
			for (n = names; n != NULL; n = n->next)
				if (same_string(n->name, d->src))
					break;
			if (n == NULL) {
				n = ckalloc(sizeof(struct name_rec));
				n->name = copy_string(d->src);
				n->next = m->next;
				n->follows = NULL;
				m->next = n;
				++nrow;
			}
			for (e = m->follows; e != NULL; e = e->next)
				if (same_string(e->name->name, n->name))
					break;
			if (e == NULL) {
				e = ckalloc(sizeof(struct edge));
				e->name = n;
				e->next = m->follows;
				m->follows = e;
			}
		}
	}
	ncol += seq_len - next_pos;

/*
for (m = names; m != NULL; m = m->next) {
printf("%s:", m->name);
for (e = m->follows; e != NULL; e = e->next)
printf(" %s", e->name->name);
putchar('\n');
}
*/

	for (n = names; n != NULL; n = n->next)
		n->color = WHITE;
	next_row_name = nrow;
	row_name = ckalloc(nrow*sizeof(char *));
	dfs(names);
	if (next_row_name != 0)
		fatal("not enough row names");
/*
for (i = 0; i < nrow; ++i)
printf("%d: %s\n", i, row_name[i]);
return 0;
*/

	row = ckalloc(nrow*sizeof(char *));
	for (i = 0; i < nrow; ++i)
		row[i] = ckalloc((ncol+1)*sizeof(char));
	col = next_pos = 0;
	for (a = mf->alignments; a != NULL; a = a->next) {
		c = a->components;
		// put in the gap before the start of the local alignment
		for (j = next_pos; j < c->start; ++j, ++col) {
			row[0][col] = s[j];
			for (i = 1; i < nrow; ++i)
				row[i][col] = '-';
		}
		// put in the local alignment
		for (i = 0; i < nrow; ++i) {
			if (c != NULL && same_string(c->src, row_name[i])) {
				for (j = 0; j < a->textSize; ++j)
					row[i][col+j] = c->text[j];
				c = c->next;
			} else {
				for (j = 0; j < a->textSize; ++j)
					row[i][col+j] = '-';
			}
		}
		if (c != NULL)
			fatal("didn't cover all rows of an alignment");
		c = a->components;
		next_pos = c->start + c->size;
		col += a->textSize;
	}
	// put in the gap after the last local alignment
	for (j = 0; j < ncol - col; ++j)
		row[0][col+j] = s[next_pos+j];
	for (i = 1; i < nrow; ++i)
		for (j = col; j < ncol; ++j)
			row[i][j] = '-';

	// print the alignment
	if (fasta) {
		for (i = 0; i < nrow; ++i) {
			row[i][ncol] = '\0';
			printf(">%s\n%s\n", row_name[i], row[i]);
		}
	} else {
		printf("%d %d\n", nrow, ncol);
		for (i = 0; i < nrow; ++i)
			printf("%s\n", row_name[i]);
		for (i = 0; i < nrow; ++i) {
			row[i][ncol] = '\0';
			printf("%s\n", row[i]);
		}
	}
	return 0;
}
