/* maf_porder -- check or enforce that an blockset meets the following "partial
*  order" condition.  We say that block A precedes block B if there is
*  a sequence S such that A and B have rows from S where A's row precedes B's
*  row in S.  We say that an blockset is partially ordered if it does not
*  contain a cycle A1, A2, ..., An of block such that Ai precedes A(i+1)
*  for 1 <= i < n, and An precedes A1.
*
*  To check for conformance with this rule, the command syntax is:
*	maf_porder - maf-file > cycle.out
*  If maf-file contains cycles, then one cycle will be left in cycle.out;
*  otherwise, cycle.out will be empty. If maf-file is partially ordered, then
*  the command
*	maf_porder maf-file > po-file
*  produces a "topological sort" of the blocks, i.e., if A immediately
*  precedes B, then A will appear before B in po-file.
*
*  Alternatively, suppose maf-file has been created by merging two partially
*  ordered maf files, e.g., using one or several calls to multiz with pairwise
*  blocks that connect the two blocksets.  Let the
*  sequences in the first blockset be seqA1, seqA2, .., seqAm and let
*  sequences in the second blockset seqB1, seqB2, .., seqBn.
*  Then the command:`
*     maf_porder maf-file seqA1 seqA2 .. seqAm + seqB1 segB2 .. SeqBn > outfile
*  (note the "+" separating the two lists of sequence names) "deconstructs"
*  blocks (removes rows from seqB1 .. seqBn and puts them in their own
*  block) in every cycle that it finds.
*
*  Implementation note: rather than dealing with all pairs (A,B) where A
*  precedes B, we work with pairs where A immediately precedes B, in the sense
*  that there is a position p in a sequence S such that A has a row from S that
*  ends p and B has a row of S that starts at p+1.  Any cycle relative to
*  "immediately precedes" is a cycle relative to "precedes" and, conversely,
*  by adding blocks if necessary any cycle relative to "precedes" can be
*  transformed to a cycle relative to "immediately precedes".  Thus freedom
*  from cycles is equivalent for the two relations.
*/
#include "lib.h"
#include "maf.h"
#include "mz_scores.h"
#include "maf_split.h"

struct mafNodeRec {
	int node;
	struct mafNodeRec *next;
};

#define WHITE 0
#define GRAY 1
#define BLACK 2

struct mafNode {
	int before, color;
	struct mafAli *maf;
	struct mafNodeRec *follows;
};

static struct mafNode *Node;
static struct mafAli **mafOut;
static int *dfs_stack, dfs_top = 0, top = 0, nseq, ntop;
char **source;

// deconstruct a block if it is possible to do so
static void maybe_split(struct mafAli *a) {
	struct mafAli *b;
	struct mafComp *c;
	int i, has_top, has_bot;

	has_top = has_bot = 0;
	for (c = a->components; c != NULL; c = c->next) {
		for (i = 0; i < nseq; ++i)
			if (same_string(c->src, source[i]))
				break;
		if (i == nseq)
			fatalf("cannot find src %s", c->src);
		if (i < ntop)
			has_top = 1;
		else
			has_bot = 1;
	}
	if (has_top && has_bot) {
		b = split(a, source, ntop, nseq);
		b->next = a->next;
		a->next = b;
	}
}

// process a cycle
static void cycle(int start_cycle) {
	int bot;

	// find first node in the cycle
	for (bot = dfs_top - 1; bot >= 0; bot--)
		if (dfs_stack[bot] == start_cycle)
			break;
	if (bot < 0)
		fatal("dfs stack underflow");

	// if only checking for a cycle, print it and die
	if (nseq <= 0) {
		for ( ; bot < dfs_top; ++bot)
			mafWrite(stdout, Node[dfs_stack[bot]].maf);
		fatal("loop");
	}

	// deconstruct nodes of the cycle
	for ( ; bot < dfs_top; ++bot)
		maybe_split(Node[dfs_stack[bot]].maf);
}

// depth-first search
static void dfs() {
	int n, c, k;
	struct mafNodeRec *m;

	while (dfs_top > 0) {
		n = dfs_stack[dfs_top-1];
		m = Node[n].follows;
		if (m == NULL) {
			Node[n].color = BLACK;
			mafOut[top++] = Node[n].maf;
			--dfs_top;
		} else {
			Node[n].follows = m->next;
			k = m->node;
			// free m ?
			if ((c = Node[k].color) == GRAY)
				cycle(k);
			else if (c == WHITE) {
				dfs_stack[dfs_top++] = k;
				Node[k].color = GRAY;
			}
		}
	}
}

// topological sort routine roughly as described on pp 549-552 of Introduction
// to Algorithms (2nd Ed.) by T.H. Cormen, C.E. Leiserson, R.L. Rivest and
// C. Stein, MIT Press, 2001
static void top_sort(int nblock) {
	int i, c;

	for (i = 0; i < nblock; ++i)
		if ((c = Node[i].color) == GRAY)
			fatal("unexpected gray node");
		else if (c == WHITE) {
			if (dfs_top != 0)
				fatalf("dfs stack confused");
			dfs_stack[dfs_top++] = i;
			Node[i].color = GRAY;
			dfs();
		}
}

/* A compRec records a beginning/ending of a component of an alignment.
*  The index of the node for the alignnment is n.  end = 0 means beginning;
*  end = 1 means that pos is the position just beyond the ending, i.e., the
*  beginning of the next component of that sequence.  Thus, we can use hashing
*  since we're searching for exact matches.
*/
struct compRec {
	int n, end, pos;
	char *src;
	struct compRec *next;
};

#define HASH_SIZE 111111

static struct compRec *bucket[HASH_SIZE];

// compute a hash value based on the sequence name and position
static int hash(char* s, int pos) {
	int x;

	for (x = pos; *s != '\0'; ++s)
		x += *s;
	return x % HASH_SIZE;
}

// the from-th node immediately precedes the to-th node
static void add_edge(int from, int to) {
	struct mafNodeRec *m;

	// maybe this is old news
	for (m = Node[from].follows; m != NULL; m = m->next)
		if (m->node == to)
			return;

	m = ckalloc(sizeof(struct mafNodeRec));
	m->node = to;
	m->next = Node[from].follows;
	Node[from].follows = m;
	Node[to].before++;
}

// process one end of a component
static void compEnd(int end, char *src, int pos, int n) {
	struct compRec *c, *new;
	int hval = hash(src, pos);

	for (c = bucket[hval]; c != NULL; c = c->next)
		if (c->pos == pos && same_string(c->src, src))
			break;
	if (c != NULL) {	// We've already seen this this seq,pos
		if (c->end == end)
			fatalf("duplication: %s:%d end = %d", src, pos, end);
		if (end)
			add_edge(n, c->n);
		else
			add_edge(c->n, n);
	}
	// store info in the hash table
	new = ckalloc(sizeof(struct compRec));
	new->end = end;
	new->src = copy_string(src);
	new->pos = pos;
	new->n = n;
	new->next = bucket[hval];
	bucket[hval] = new;
}

int main(int argc, char **argv) {
	struct mafFile *mf;
	struct mafAli *a, *A, *anext;
	struct mafComp *c;
	int i, j, nblock, seen_plus, print_mafs = 1;

	argv0 = "maf_porder";
	if (argc < 2)
		fatal("args = [-] maf-file [seqA1 .. seqAm + seqB1 ... seqBn");
	if (same_string(argv[1], "-")) {
		print_mafs = 0;
		--argc;
		++argv;
	}
	if ((nseq = argc - 3) > 0) {
		source = ckalloc(nseq*sizeof(char *));
		ntop = seen_plus = 0;
		for (j = 0, i = 2; i < argc; ++i)
			if (same_string(argv[i], "+")) {
				if (ntop == 0)
					fatal("saw + immediately");
				seen_plus = 1;
			} else {
				if (!seen_plus)
					++ntop;
				if (j >= nseq)
					fatal("bad command args");
				source[j++] = copy_string(argv[i]);
			}
		if (!seen_plus)
			fatal("needs a + in command line");
		init_scores70();
	}

	mf = mafReadAll(argv[1], 1);
	A = mf->alignments;
	for (nblock = 0, a = A; a != NULL; a = a->next, ++nblock)
		;
	Node = ckalloc(nblock * sizeof(struct mafNode));
	mafOut = ckalloc(nblock * sizeof(struct mafAli *));
	dfs_stack = ckalloc(nblock * sizeof(int));
	for (i = 0, a = A; a != NULL; a = anext, ++i) {
		anext = a->next;
		a->next = NULL;
		Node[i].maf = a;
		Node[i].follows = NULL;
		Node[i].color = WHITE;
		Node[i].before = 0;
		for (c = a->components; c != NULL; c = c->next) {
			compEnd(0, c->src, c->start, i);
			compEnd(1, c->src, c->start + c->size, i);
		}
	}
// code to print the directed graph
/*
for (i = 0; i < nblock; ++i) {
struct mafNodeRec *z;
printf("%d:", i);
for (z = Node[i].follows; z != NULL; z = z->next)
printf(" %d", z->node);
printf("\n");
mafWrite(stdout, Node[i].maf);
}
exit(0);
*/
	// do some error checking
	for (i = 0; i < nblock; ++i) {
		if (Node[i].before == 0)
			for (c = Node[i].maf->components; c != NULL;
			     c = c->next)
				if (c->start != 0) {
					mafWrite(stderr, Node[i].maf);
					fatalf("has no predecessors");
				}
		if (Node[i].follows == NULL)
			for (c = Node[i].maf->components; c != NULL;
			     c = c->next)
				if (c->start + c->size != c->srcSize) {
					mafWrite(stderr, Node[i].maf);
					fatalf("has no successors");
				}
	}

	top_sort(nblock);
	if (print_mafs) {
		mafWriteStart(stdout, argv0);
		while (top > 0)
			for (a = mafOut[--top]; a != NULL; a = a->next)
				mafWrite(stdout, a);
	}
	
	return 0;
}
